.NOLIST
    ; -----------------------------------------------
    ; DOS handles for standard input/output devices.
    ; -----------------------------------------------

    HSTDIN   EQU 0
    HSTDOUT  EQU 1

    ; -----------------------------------------------------
    ; This is the MASM32 reparg macro with *no* changes.
    ;
    ; "This macro replaces quoted text with a DATA section
    ; OFFSET and returns it in ADDR "name" format."
    ; -----------------------------------------------------

    reparg MACRO arg
      LOCAL nustr
      quot SUBSTR <arg>, 1, 1
      IFIDN quot,<">          ;; if 1st char = "
        .data
          nustr db arg, 0     ;; write arg to .DATA section
        .code
        EXITM <ADDR nustr>    ;; append name to ADDR operator
      ELSE
        EXITM <arg>           ;; else return arg
      ENDIF
    ENDM

    ; -----------------------------------------------------
    ; This is a VARARG version of the MASM32 print macro.
    ; It will accept either a quoted string, or the offset
    ; address of a null-terminated string in the data
    ; segment.
    ; -----------------------------------------------------

    print MACRO varname:VARARG
      IFDEF DEBUG_ON
        pushf
        pushad
      ENDIF
      FOR arg, <varname>
        invoke stdout, reparg(arg)
      ENDM
      IFDEF DEBUG_ON
        popad
        popf
      ENDIF
    ENDM

    ; ------------------------------------------------------
    ; This is the MASM32 chr$ macro with *no* changes.
    ;
    ; "This macro takes quoted string and ascii numbers in
    ; the same format as a .DATA section entry. It returns
    ; the offset in the .DATA section of the data contained
    ; in its parameters."
    ; ------------------------------------------------------

    chr$ MACRO any_text:VARARG
      LOCAL txtname
      .data
        txtname db any_text,0
      .code
      EXITM <OFFSET txtname>
    ENDM

    ; ----------------------------------------------------
    ; These macros make it easier to print signed decimal
    ; and unsigned hex values.
    ;
    ; Note that the decimal conversion procedures cannot
    ; correctly handle out of range values.
    ;
    ; Note also that the conversions reuse a single output
    ; buffer, so if more then one conversion is placed in
    ; a single print statement all of the conversions will
    ; effectively return the value of the last conversion.
    ;
    ; The pushs and pops in the byte versions are because
    ; for a byte register argument other than AL invoke
    ; uses AL for a scratch register.
    ; ----------------------------------------------------

    byte$ MACRO src
      push ax
      invoke b2dec,src,ADDR outBuf
      pop ax
      EXITM <ADDR outBuf>
    ENDM

    hexbyte$ MACRO src
      push ax
      invoke b2hex,src,ADDR outBuf
      pop ax
      EXITM <ADDR outBuf>
    ENDM

    word$ MACRO src
      invoke w2dec,src,ADDR outBuf
      EXITM <ADDR outBuf>
    ENDM

    hexword$ MACRO src
      invoke w2hex,src,ADDR outBuf
      EXITM <ADDR outBuf>
    ENDM

    dword$ MACRO src
      invoke dw2dec,src,ADDR outBuf
      EXITM <ADDR outBuf>
    ENDM

    hexdword$ MACRO src
      invoke dw2hex,src,ADDR outBuf
      EXITM <ADDR outBuf>
    ENDM

    ; -------------------------------------------------
    ; This macro makes it easier to print characters.
    ; AX must be preserved around invoke because for
    ; a byte register argument other than AL invoke
    ; uses AL for a scratch register.
    ; -------------------------------------------------

    putch MACRO char
      push ax
      invoke putchar, char
      pop ax
    ENDM

    cls MACRO
      call _cls
    ENDM

    ; ----------------------------------------------------
    ; This macro sets the cursor position for display
    ; page 0 to the specified row and column coordinates.
    ; The unusual feature is that either or both of the
    ; coordinate parameters may be left blank, causing
    ; the cursor position with respect to the blank
    ; coordinate to remain unchanged.
    ; ----------------------------------------------------

    loc MACRO row,col
      IFNB <row>
        IFNB <col>
          invoke locate, row, col
        ELSE
          call cursorcol
          invoke locate, row, al
        ENDIF
      ELSE
        IFNB <col>
          call cursorrow
          invoke locate, al, col
        ELSE
          call cursorrow
          mov   ah, al
          call cursorcol
          invoke locate, ah, al
        ENDIF
      ENDIF
    ENDM

; -------------------------------------------------------------

.data
    inbuf       db 130 dup(0)
    outbuf      db 30 dup(0)
    crlf        db 13, 10, 0
    NL          dw offset crlf
    backspace   db 8, 0
    BS          dw OFFSET backspace
.code

; -------------------------------------------------------------
; This proc calls the Read File or Device function to read
; up to <blen> bytes from STDIN and store them in the
; specified buffer.
; -------------------------------------------------------------

stdin proc c uses bx cx dx npdest:WORD, blen:WORD
    mov bx, HSTDIN
    mov cx, blen
    mov dx, npdest
    mov ah, 3fh
    int 21h
    mov bx, npdest
  @@:                         ; Replace CR with null
    cmp BYTE PTR [bx], 0
    je  @F
    cmp BYTE PTR [bx], 13
    je  @F
    inc bx
    jmp @B
  @@:
    mov BYTE PTR [bx], 0
    ret
stdin endp

; -------------------------------------------------------------
; This proc returns the length of the specified
; null-terminated string in AX.
; -------------------------------------------------------------

szlen proc c uses bx npsz:WORD
    mov bx, npsz
  runLoop:
    cmp BYTE PTR [bx], 0      ; Compare bytes until find null
    jz  @F
    inc bx
    jmp runLoop
  @@:
    mov ax, bx                ; Calc length and return in AX
    sub ax, npsz
    ret
szlen endp

; -------------------------------------------------------------
; This proc calls the Write File or Device function to
; output the specified null-terminated string to STDOUT.
; -------------------------------------------------------------

stdout proc c uses ax bx cx dx npsz:WORD
    invoke szlen, npsz
    mov cx, ax
    mov bx, HSTDOUT
    mov dx, npsz
    mov ah, 40h
    int 21h
    ret
stdout endp

; -------------------------------------------------------------
; This proc calls the Display Character function to output
; the specified character to STDOUT.
; -------------------------------------------------------------

putchar proc c uses ax dx char:BYTE
    mov ah, 2
    mov dl, char
    int 21h
    ret
putchar endp

; -------------------------------------------------------------
; This proc calls the BIOS Read Keyboard Status and Read
; Keyboard Input functions to flush the keyboard buffer,
; wait for a key press, and then flush the buffer again,
; leaving it empty. The BIOS scan code for the key pressed
; is returned in AH and the character code in AL.
; -------------------------------------------------------------

waitkey proc c
  @@:
    mov ah, 1
    int 16h                   ; Check for key press
    jz  @F                    ; Jump if buffer empty
    mov ah, 0
    int 16h                   ; Get key
    jmp @B                    ; Repeat
  @@:

    ;mov ax, 1680h             ; MS-DOS Idle Call
    ;int 2fh

    mov ah, 1
    int 16h                   ; Check for key press
    jz  @B                    ; Repeat
    mov ah, 0
    int 16h                   ; Get key
    push ax                   ; Preserve it
  @@:
    mov ah, 1
    int 16h                   ; Check for key press
    jz  @F                    ; Jump if buffer empty
    mov ah, 0
    int 16h                   ; Get key
    jmp @B                    ; Repeat
  @@:
    pop ax                    ; Recover key
    ret
waitkey endp

; -------------------------------------------------------------
; This proc calls the BIOS Set Cursor Position function to
; set the cursor position for display page 0 to <row>,<col>
; (upper left corner is 0,0).
; -------------------------------------------------------------

locate proc c uses ax bx dx row:BYTE, col:BYTE
    mov ah, 2
    mov bh, 0                 ; Display page
    mov dh, row
    mov dl, col
    int 10h
    ret
locate endp

; -------------------------------------------------------------
; These procs return the row and column cursor positions
; for display page 0 in AL.
; -------------------------------------------------------------

cursorrow proc c uses es
    push 40h                  ; Set ES to the BIOS data area
    pop es
    mov al, es:[51h]          ; Read the appropriate byte
    ret
cursorrow endp

cursorcol proc c uses es
    push 40h
    pop es
    mov al, es:[50h]
    ret
cursorcol endp

; -------------------------------------------------------------
; This proc clears the screen by calling the BIOS Scroll
; Current Page Up function with a scroll (row) count of zero,
; and sets the cursor position to the upper left corner of
; the screen. An 80x25 text mode screen is assumed.
; -------------------------------------------------------------

_cls proc c uses ax bx cx dx
    mov ah, 6                 ; Function number
    mov al, 0                 ; Row count
    mov bh, 7                 ; Attribute for blanked lines
                              ; (white on black)
    mov ch, 0                 ; Top row of window
    mov cl, 0                 ; Left-most column of window
    mov dh, 24                ; Bottom row of window
    mov dl, 79                ; Right-most column of window
    int 10h
    invoke locate, 0, 0
    ret
_cls endp

; -------------------------------------------------------------
; This proc converts the specified null-terminated decimal
; string to a signed 32-bit integer value and returns the
; result in EAX. The procedure stops reading the string at the
; first non-numeric character, which may be the terminating
; null. This procedure cannot correctly handle values outside
; the range -2147483648 to 2147483647.
; -------------------------------------------------------------

atodw proc c uses dx si npsz:WORD
    LOCAL fNeg:WORD

    mov si, npsz
    xor edx, edx              ; Use EDX to store value
  @@:
    lodsb
    cmp al, 32
    je  @B                    ; Scan off leading spaces
    cmp al, 9
    je  @B                    ; Scan off leading tabs
    cmp al, '+'
    je  @B                    ; Scan off leading '+'
    cmp al, '-'               ; Check for leading '-'
    jne @F
    mov fNeg, 1
    jmp digitLoop
  @@:
    mov fNeg, 0               ; Char is digit, clear flag and
    dec si                    ;  decrement SI to repeat char
  digitLoop:
    lodsb
    cmp al, '0'
    jb  endOfString           ; Stop at non-numeric char
    cmp al, '9'
    ja  endOfString           ; Stop at non-numeric char
    cmp edx, 0
    jne @F
    cmp al, '0'
    je  digitLoop             ; Ignore any leading zeros
  @@:
    sub al, '0'               ; Convert to decimal value
    cmp edx, 0
    je  @F                    ; If not first digit
    imul edx, 10              ;  multiply total times base
  @@:
    movzx eax, al             ; Zero extend AL into EAX
    add edx, eax              ; Add in value for digit
    jmp digitLoop
  endOfString:
    mov eax, edx
    cmp fNeg, 0
    je  @F
    neg eax
  @@:
    ret
atodw endp

; -------------------------------------------------------------
; This proc converts a byte to a signed null-terminated
; decimal string and stores the result (4 characters max)
; in the specified buffer.
;
; This proc cannot correctly handle values that are outside
; the range -128 to 127.
; -------------------------------------------------------------

b2dec proc c uses ax bx cx dx di src:BYTE, npdest:WORD
    mov al, src
    mov di, npdest
    test al, al               ; Test for negative value
    jns @F                    ; Make positive if negative
    neg al
    mov BYTE PTR [di], '-'    ; Store sign in buffer
    inc di                    ; Adjust DI for next character
  @@:
    mov bl, 10                ; Init divisor
    xor cx, cx                ; Zero digit counter
  divLoop:
    xor ah, ah                ; Zero MSB of dividend
    div bl                    ; Divide value by 10
    movzx dx, ah
    push dx                   ; Push remainder
    inc cx                    ; Increment digit count
    or  al, al                ; Repeat if quotient > 0
    jnz divLoop
  storeLoop:
    pop ax                    ; Pop digit off stack
    add al, '0'               ; Convert to ascii
    mov [di], al              ; Store in buffer
    inc di                    ; Adjust DI for next character
    loop storeLoop            ; Repeat CX times
    mov BYTE PTR [di], 0      ; Append trailing null
    ret
b2dec endp

; -------------------------------------------------------------
; This proc converts a byte to a null-terminated 2-character
; hex string and stores the result in the specified buffer.
; -------------------------------------------------------------

  .data
    hexChars db "0123456789ABCDEF"
  .code

b2hex proc c uses ax bx dx di src:BYTE, npdest:WORD
    mov dl, src
    mov di, npdest
    add di, 1                 ; Set DI to end of buffer
    mov BYTE PTR [di+1], 0    ; Append trailing null
  @@:
    movzx bx, dl              ; Copy current value into BX
    and bx, 0Fh               ; Mask off all but lower nibble
    mov al, hexChars[bx]      ; Load corresponding char into AL
    mov [di], al              ; Store in buffer
    shr dl, 4                 ; Shift next nibble into position
    dec di                    ; Adjust DI for next character
    cmp di, npdest            ; Repeat for all 2 characters
    jnb @B
    ret
b2hex endp

; -------------------------------------------------------------
; This proc converts a word to a signed null-terminated
; decimal string and stores the result (6 characters max)
; in the specified buffer.
;
; This proc cannot correctly handle values that are outside
; the range -32768 to 32767.
; -------------------------------------------------------------

w2dec proc c uses ax bx cx dx di src:WORD, npdest:WORD
    mov ax, src
    mov di, npdest
    test ax, ax               ; Test for negative value
    jns @F                    ; Make positive if negative
    neg ax
    mov BYTE PTR [di], '-'    ; Store sign in buffer
    inc di                    ; Adjust DI for next character
  @@:
    mov bx, 10                ; Init divisor
    xor cx, cx                ; Zero digit counter
  divLoop:
    xor dx, dx                ; Zero MSW of dividend
    div bx                    ; Divide value by 10
    push dx                   ; Push remainder
    inc cx                    ; Increment digit count
    or  ax, ax                ; Repeat if quotient > 0
    jnz divLoop
  storeLoop:
    pop ax                    ; Pop digit off stack
    add al, '0'               ; Convert to ascii
    mov [di], al              ; Store in buffer
    inc di                    ; Adjust DI for next character
    loop storeLoop            ; Repeat CX times
    mov BYTE PTR [di], 0      ; Append trailing null
    ret
w2dec endp

; -------------------------------------------------------------
; This proc converts a word to a null-terminated 4-character
; hex string and stores the result in the specified buffer.
; -------------------------------------------------------------

w2hex proc c uses ax bx dx di src:WORD, npdest:WORD
    mov dx, src
    mov di, npdest
    add di, 3                 ; Set DI to end of buffer
    mov BYTE PTR [di+1], 0    ; Append trailing null
  @@:
    mov bx, dx                ; Copy current value into BX
    and bx, 0Fh               ; Mask off all but lower nibble
    mov al, hexChars[bx]      ; Load corresponding char into AL
    mov [di], al              ; Store in buffer
    shr dx, 4                 ; Shift next nibble into position
    dec di                    ; Adjust DI for next character
    cmp di, npdest            ; Repeat for all 4 characters
    jnb @B
    ret
w2hex endp

; -------------------------------------------------------------
; This proc converts a dword to an signed null-terminated
; decimal string and stores the result (11 characters max)
; in the specified buffer.
;
; This proc cannot correctly handle values that are outside
; the range -2147483648 to 2147483647.
; -------------------------------------------------------------

dw2dec proc c uses ax bx cx dx di src:DWORD, npdest:WORD
    mov eax, src
    mov di, npdest
    test eax, eax             ; Test for negative value
    jns @F
    neg eax                   ; Make positive if negative
    mov BYTE PTR[di], '-'     ; Store sign in buffer
    inc di
  @@:
    mov ebx, 10               ; Init divisor
    xor cx, cx                ; Zero digit counter
  divLoop:
    xor edx, edx              ; Zero MSDW of dividend
    div ebx                   ; Divide value by 10
    push dx                   ; Push remainder
    inc cx                    ; Increment digit count
    or  eax, eax              ; Repeat if quotient > 0
    jnz divLoop
  storeLoop:
    pop ax                    ; Pop digit off stack
    add al, '0'               ; Convert to ascii
    mov [di], al              ; Store in buffer
    inc di                    ; Adjust DI for next character
    loop storeLoop            ; Repeat CX times
    mov BYTE PTR [di], 0      ; Append trailing null
    ret
dw2dec endp

; -------------------------------------------------------------
; This proc converts a dword to a null-terminated
; 8-character hex string and stores the result in the
; specified buffer.
; -------------------------------------------------------------

dw2hex proc c uses ax bx dx di src:DWORD, npdest:WORD
    mov edx, src
    mov di, npdest
    add di, 7                 ; Set DI to end of buffer
    mov BYTE PTR [di+1], 0    ; Append trailing null
  @@:
    mov ebx, edx              ; Copy current value into EBX
    and ebx, 0Fh              ; Mask off all but lower nibble
    mov al, hexChars[ebx]     ; Load corresponding char into AL
    mov [di], al              ; Store in buffer
    shr edx, 4                ; Shift next nibble into position
    dec di                    ; Adjust DI for next character
    cmp di, npdest
    jnb @B                    ; Repeat for all 8 characters
    ret
dw2hex endp

; -------------------------------------------------------------
; This proc delays for the specified number of memory refresh
; requests. The requests are generated by PIT channel 1,
; normally programmed by the BIOS to generate an output once
; every 18/1193182 = 15.09 microseconds. Each request toggles
; bit 4 of I/O port 61h.
; -------------------------------------------------------------

delay proc uses ax ecx dwcount:DWORD
    mov ecx, dwcount
  waituntilset:
    in  al, 61h
    test al, 10h
    jz  waituntilset
  waituntilclear:
    in  al, 61h
    test al, 10h
    jnz waituntilclear
    sub ecx, 1
    jnz waituntilset
    ret
delay endp

; -------------------------------------------------------------
.LIST