;====================================================================
; Much of this code is adapted from Geoff Chappell's book
; DOS Internals (first edition), Addison-Wesley 1994.
; ISBN 0-201-60865-9.
;====================================================================
.model small, c
.386p
;====================================================================

SEGMENT_DESCRIPTOR struc
    seg_limit       dw  0
    seg_base        dw  0
    base_mid        db  0
    access_rights   db  0
    extra_byte      db  0
    base_high       db  0
SEGMENT_DESCRIPTOR ends

TABLE_DESCRIPTOR struc
    limit           dw  0
    base            dd  0
TABLE_DESCRIPTOR ends

;====================================================================
include support.asm
;====================================================================
.stack

.data

    ;----------------------------------------------------------
    ; Create and initialize a temporary GDT. The Null selector
    ; is first, followed by a descriptor with a base of zero,
    ; a maximum segment limit, and page granularity (so the
    ; segment covers the entire 32-bit address space).
    ;  Segment limit:  0fffffh (max assuming page granularity)
    ;  Segment base:   0
    ;  Access rights:  P=1,DPL=0,S=1,E=0,ED=0,W=1,A=1
    ;  Extra byte:     G=1,D=0,limit19-16 = 0fh
    ;----------------------------------------------------------

    gdt SEGMENT_DESCRIPTOR  <>, < 0ffffh, , , 93h, 8fh >

    ;------------------------------------------------------
    ; Define our selector value (effectively a GDT index).
    ;------------------------------------------------------

    GS_selector = sizeof SEGMENT_DESCRIPTOR

    ;---------------------------------------------------
    ; Create a GDT descriptor and initialize the limit.
    ;---------------------------------------------------

    GS_descriptor TABLE_DESCRIPTOR <sizeof gdt - 1>

.code

;====================================================================
;-----------------------------------------------------------
; This procedure steps up to PM temporarily to load the GS
; segment descriptor cache, setting the segment limit and
; granularity for a 4GB segment size, and the segment base
; to 0.
;
; After the return to RM the new limit and base will remain
; in effect up to the point that a new value is loaded into
; the register. The reason for using GS here is that the
; DOS and BIOS functions that this code calls are unlikely
; to use it.
;-----------------------------------------------------------

SetGSLimit proc uses es

    ;------------------------------------
    ; Set GDT base address and load GDT.
    ;------------------------------------

    mov	ax, ds
    movzx	eax, ax
    shl	eax, 4
    add	eax, offset gdt
    mov	GS_descriptor.base, eax
    mov	bx, GS_selector
    pushf
    cli
    lgdt fword ptr GS_descriptor

    ;-----------------------
    ; Enter protected mode.
    ;-----------------------

    mov	eax, cr0
    or  al, 01h
    mov cr0, eax

    ;----------------------------------------------------------------
    ; Flush instruction queue (discard instructions decoded for RM).
    ;----------------------------------------------------------------

    jmp	@f
  @@:

    ;----------------------------
    ; Load the descriptor cache.
    ;----------------------------

    mov	gs, bx

    ;----------------------
    ; Return to real mode.
    ;----------------------

    and	al, not 01h
    mov	cr0, eax

    ;----------------------------------------------------------------
    ; Flush instruction queue (discard instructions decoded for PM).
    ;----------------------------------------------------------------

    jmp	@f
  @@:

    popf
    ret

SetGSLimit endp

;====================================================================

EnableA20_PS2 proc

    cli
    in  al, 92h
    or  al, 2
    out 92h, al
    sti
    ret

EnableA20_PS2 endp

;====================================================================

;----------------------------------------------------------
; Returns non-zero if the A20 address line is enabled.
; Assumes the A20 address line is disabled if the
; address FFFF:10 wraps to the beginning of memory
; (so the words at 0:0 and FFFF:10 are identical).
;----------------------------------------------------------
A20Enabled proc uses ds es

    cli
    push  0ffffh
    pop   es
    mov   di, 10h           ; ES:DI = FFFF:10
    push  0
    pop   ds
    mov   si, 0             ; DS:SI = 0:0
    not   WORD PTR ds:[si]  ; Modify value at 0:0
    xor   ax, ax
    cld
    cmpsw
    je    @F                ; Jump if A20 disabled
    inc   ax
  @@:
    not   WORD PTR ds:[si]  ; Restore value at 0:0
    sti
    ret

A20Enabled endp

;====================================================================
.startup
;====================================================================

    mov ax, 0f000h
    mov gs, ax
    mov edx, 0fff0h
    xor ecx, ecx
    .while ecx < 16
      mov al, gs:[edx+ecx]
      print hexbyte$(al)
      inc ecx
    .endw
    print chr$(13,10)

    call A20Enabled
    print word$(ax),NL

    call EnableA20_PS2

    call A20Enabled
    print word$(ax),NL

    call SetGSLimit

    cli

    mov edx, 0ffff0h
    xor ecx, ecx
    .while ecx < 16
      mov al, gs:[edx+ecx]
      print hexbyte$(al)
      inc ecx
    .endw
    print chr$(13,10)

    mov edx, 0fffffff0h
    xor ecx, ecx
    .while ecx < 16
      mov al, gs:[edx+ecx]
      print hexbyte$(al)
      inc ecx
    .endw
    print chr$(13,10)

    sti

    call waitkey

;====================================================================
.exit
end
